import concurrent.futures
import math
from multiprocessing import Manager
from os import PathLike
from pathlib import Path

from rich.progress import Progress

from plateaukit import parallel, utils
from plateaukit.logger import logger

from ._serial import geojson_from_gml_serial_with_quit


def _geojson_from_citygml(
    infiles,
    outfile,
    *,
    codelist_infiles,
    types,
    split,
    zipfile,
    seq: bool,
    progress={},
    simple_output=False,
    **kwargs,
):
    group_size = math.ceil(len(infiles) / split)
    logger.debug(f"GMLs per GeoJSON: {group_size}")
    infile_groups = utils.chunker(infiles, group_size)

    # https://www.deanmontgomery.com/2022/03/24/rich-progress-and-multiprocessing/
    with Progress() as rprogress:
        overall_progress_description = progress.get("description", "Processing...")
        overall_task_id = rprogress.add_task(description=overall_progress_description)

        with Manager() as manager:
            _quit = manager.Event()
            _progress = manager.dict()
            with concurrent.futures.ProcessPoolExecutor(max_workers=None) as pool:
                futures = []
                futures_status = dict()
                for i, infile_group in enumerate(infile_groups):
                    stem = Path(outfile).stem
                    if split > 1:
                        group_outfile = Path(outfile).with_stem(f"{stem}.{i + 1}")
                    else:
                        group_outfile = outfile
                    if simple_output:
                        task_id = None
                    else:
                        task_id = rprogress.add_task(
                            f"[cyan]Progress #{i + 1}", total=len(infile_group)
                        )
                    future = pool.submit(
                        geojson_from_gml_serial_with_quit,
                        infile_group,
                        group_outfile,
                        codelist_infiles=codelist_infiles,
                        types=types,
                        seq=seq,
                        zipfile=zipfile,
                        task_id=task_id,
                        _progress=_progress,
                        _quit=_quit,
                        **kwargs,
                    )
                    futures.append(future)
                    futures_status[future] = {
                        "task_id": task_id,
                        "counter": i + 1,
                        "failed": False,
                    }

                parallel.wait_futures(
                    futures,
                    pool,
                    _quit,
                    futures_status,
                    overall_progress={
                        "task_id": overall_task_id,
                        "description": overall_progress_description,
                    },
                    rich_progress=rprogress,
                    shared_progress_status=_progress,
                )


def geojson_from_citygml(
    infiles: list[str | PathLike],
    outfile: str | PathLike,
    *,
    types: list[str],
    altitude: bool = True,
    seq: bool = False,
    split: int,
    include_type: bool = False,
    zipfile: str | PathLike | None = None,
    codelist_infiles=None,
    **kwargs,
):
    """Generate GeoJSON file(s) from CityGML files."""

    import glob

    if zipfile is None:
        expanded_infiles = []
        for infile in infiles:
            expanded_infiles.extend(glob.glob(infile))

        infiles = expanded_infiles

    infiles = sorted(infiles)

    stem = Path(outfile).stem

    for type in types:
        if len(types) > 1:
            type_outfile = Path(outfile).with_stem(f"{stem}.{type}")
        else:
            type_outfile = outfile

        if type == "bldg":
            """建築物、建築物部分、建築物付属物、及びこれらの境界面"""
            _geojson_from_citygml(
                infiles,
                type_outfile,
                codelist_infiles=codelist_infiles,
                types=["Building"],
                include_type=include_type,
                seq=seq,
                split=split,
                lod=[0],
                altitude=altitude,
                allow_geometry_collection=False,
                zipfile=zipfile,
                **kwargs,
            )
        elif type == "brid":
            """橋梁"""
            _geojson_from_citygml(
                infiles,
                type_outfile,
                codelist_infiles=codelist_infiles,
                types=["Bridge"],
                include_type=include_type,
                seq=seq,
                split=split,
                lod=[2],
                attributes=[],
                altitude=altitude,
                allow_geometry_collection=True,
                zipfile=zipfile,
                **kwargs,
            )
        elif type == "tran":
            """道路"""
            _geojson_from_citygml(
                infiles,
                type_outfile,
                codelist_infiles=codelist_infiles,
                types=["Road"],
                include_type=include_type,
                seq=seq,
                split=split,
                lod=[1],
                attributes=[],
                altitude=altitude,  # TODO: can be False
                allow_geometry_collection=True,
                zipfile=zipfile,
                **kwargs,
            )
        elif type == "dem":
            """地形(起伏)"""
            raise NotImplementedError("dem")
        elif type == "fld":
            """洪水浸水想定区域"""
            raise NotImplementedError("fld")
        elif type == "lsld":
            """土砂災害警戒区域"""
            raise NotImplementedError("lsld")
        elif type == "luse":
            """土地利用"""
            raise NotImplementedError("luse")
        elif type == "urf":
            """都市計画区域、区域区分、地域地区"""
            raise NotImplementedError("urf")
        else:
            raise NotImplementedError(type)



import io
import re
import zipfile
from os import PathLike
from pathlib import Path

from fs import open_fs
from lxml import etree

from plateaukit.config import Config
from plateaukit.formats.citygml import constants
from plateaukit.formats.citygml.parsers import CodelistParser


class CityGMLDataset:
    """A (PLATEAU) CityGML dataset.

    Attributes:
        dataset_id: Dataset ID.
    """

    dataset_id: str

    def __init__(self, dataset_id: str | PathLike):
        """Initialize a dataset.

        Args:
            dataset_id: Dataset ID
        """
        # TODO: Support zip file as input

        self.dataset_id = str(dataset_id)

        config = Config()
        record = config.datasets.get(self.dataset_id, None)

        if not record:
            raise RuntimeError("Missing dataset record")

        if "citygml" not in record:
            raise RuntimeError("Missing CityGML data")

        file_path = Path(record["citygml"])

        self.file_path = file_path

    def udx_dirs(self):
        """List UDX directories in the dataset."""

        pat = re.compile(r".*udx\/([^\/]*?)\/$")

        with zipfile.ZipFile(self.file_path) as f:
            namelist = f.namelist()
            dirs = {
                pat.match(target).group(1): str(Path("/", pat.match(target).group(0)))
                for target in namelist
                if pat.match(target)
            }

        return dirs

    def _list_udx_files(self, udx_type: str):
        """List files of a udx type in the dataset."""

        udx_dirs = self.udx_dirs()

        if udx_type not in udx_dirs:
            raise RuntimeError(f"Missing udx directory: {udx_type}")

        pat = re.compile(rf".*\/udx\/{udx_type}\/[^\/]*\.gml$")
        with zipfile.ZipFile(self.file_path) as f:
            namelist = f.namelist()
            files = list(filter(lambda x: pat.match(x), namelist))
            files = [str(Path("/", target)) for target in files]

        return files

    def scan_attributes(self, udx_type: str):
        """Scan city object properties in the dataset."""

        files = self._list_udx_files(udx_type)

        codelists = self.codelists()

        zip_fs = open_fs(f"zip://{self.file_path}")

        props = []

        for file in files:
            base_path = Path(file).parent

            with zip_fs.open(file, "rb") as f:
                # root = etree.parse(f).getroot()
                # for city_obj in root.iterfind(
                #     ".//core:cityObjectMember/*", constants.nsmap
                # ):

                tag = f"{{{constants.nsmap['core']}}}cityObjectMember"
                itertree = etree.iterparse(f, events=("end",), tag=tag)
                _, root = next(itertree)

                for _ev, co_root in itertree:
                    it = co_root.iterchildren()
                    city_obj = next(it)

                    for el in city_obj:
                        _qname = etree.QName(el)
                        ns = _qname.namespace
                        localname = _qname.localname
                        ns_prefix = constants.nsmap.inverse.get(ns, None)
                        if ns_prefix:
                            qname = f"{ns_prefix}:{localname}"
                        else:
                            qname = _qname.text

                        prop = {
                            "tag": qname,
                        }

                        if el.tag == f"{{{constants.nsmap['gen']}}}stringAttribute":
                            name = el.attrib["name"]
                            prop["name"] = name

                        elif (
                            el.tag
                            == f"{{{constants.nsmap['uro']}}}keyValuePairAttribute"
                        ):
                            for child in el.iterfind(
                                "./uro:KeyValuePairAttribute", constants.nsmap
                            ):
                                # print("child", child)
                                el_key = child.find("./uro:key", constants.nsmap)
                                key = el_key.text
                                codelist_path = el_key.attrib.get("codeSpace", None)
                                codelist_path = str(
                                    Path(base_path, codelist_path).resolve()
                                )
                                # print("codelist_path", codelist_path)
                                codelist = codelists.get(codelist_path, None)
                                # print("codelist", codelist)
                                if codelist:
                                    key_name = codelist[key]
                                    prop["name"] = key_name

                        if "name" not in prop:
                            if qname in constants.tag_display_names:
                                prop["name"] = constants.tag_display_names[qname].get(
                                    "ja", None
                                )
                            else:
                                prop["name"] = localname

                        if prop not in props:
                            props.append(prop)
                            yield prop
                            # logger.debug(f"{prop}")

                    # co_root.clear()  # NOTE: Bad performance impact
                    root.clear()  # NOTE: Unnecessary?

    def codelists(self):
        """Get codelists from the dataset."""

        file_path = self.file_path

        codelist_infiles = []

        pat = re.compile(r".*codelists\/.*\.xml$")

        with zipfile.ZipFile(file_path) as f:
            namelist = f.namelist()
            codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
            codelist_infiles = [str(Path("/", target)) for target in codelist_infiles]

        zip_fs = open_fs(f"zip://{file_path}")

        codelist_file_map = dict()

        for codelist_infile in codelist_infiles:
            if zipfile is not None:
                with zip_fs.open(codelist_infile, "rb") as f:
                    codelist_file_map[codelist_infile] = io.BytesIO(f.read())
            else:
                with open(codelist_infile, "rb") as f:
                    codelist_file_map[codelist_infile] = io.BytesIO(f.read())

        codelists = {}

        for codelist_infile, file_obj in codelist_file_map.items():
            parser = CodelistParser()
            codelists[codelist_infile] = parser.parse(file_obj)

        return codelists



import warnings

from loguru import logger
from rich.logging import RichHandler

logger = logger.bind(name="plateaukit")

logger.configure(
    handlers=[
        {
            "sink": RichHandler(markup=True),
            "format": "{message}",
            "level": "INFO",
        }
    ]
)


# https://loguru.readthedocs.io/en/stable/resources/recipes.html#capturing-standard-stdout-stderr-and-warnings
def capture_warnings():
    global warnings
    showwarning_ = warnings.showwarning

    def showwarning(message, *args, **kwargs):
        logger.debug(message)
        # showwarning_(message, *args, **kwargs)

    warnings.showwarning = showwarning


capture_warnings()


def set_log_level(level: str):
    """Set the log level.

    Args:
        level: Log level
    """
    logger.configure(
        handlers=[
            {"sink": RichHandler(markup=True), "format": "{message}", "level": level}
        ]
    )



from __future__ import annotations

import click
from rich.console import Console

from plateaukit.config import Config


@click.command("info")
@click.argument("dataset_id", nargs=1, required=True)
def info_cmd(dataset_id):
    """Show PLATEAU dataset information.

    PLATEAU データセットの情報を表示します。
    """
    from plateaukit.download import city_list
    from plateaukit.formats.citygml import CityGMLDataset, constants

    config = Config()

    remote_info = next(filter(lambda x: x["dataset_id"] == dataset_id, city_list), None)

    if not remote_info:
        raise click.UsageError(f"Unknown dataset: {dataset_id}")

    console = Console(highlight=False)

    console.print(f"[b]{dataset_id}[/b]")
    console.print(f"[u]{remote_info['homepage']}")
    console.print(f"[b]Name:[/b] {remote_info['city_name']}")
    console.print(f"[b]Version:[/b] {remote_info['version']}")

    record = config.datasets.get(dataset_id, None)

    if not record:
        console.print("[b]Installed Files:[/b]\n  (None)")
        return

    console.print("[b]Installed Files:[/b]")
    for format, path in record.items():
        console.print(f"  {format}: {path}")

    dataset = CityGMLDataset(dataset_id)
    udx_dirs = dataset.udx_dirs()

    console.print("[b]Data Types:[/b]")
    for udx_type, path in udx_dirs.items():
        type_name = constants.data_type_display_names.get(udx_type, {}).get(
            "ja", udx_type
        )
        console.print(f"  {type_name} ({udx_type}): {path}")

    console.print("[b]Attributes:[/b]")

    with console.status(
        "[bold]Processing...", spinner="simpleDotsScrolling"
    ) as _status:
        for udx_type in udx_dirs.keys():
            type_name = constants.data_type_display_names.get(udx_type, {}).get(
                "ja", udx_type
            )
            console.print(f"  [b]{type_name} ({udx_type}):[/b]")
            for attr in dataset.scan_attributes(udx_type):
                console.print(f"    {attr.get('name')} ({attr['tag']})")



from plateaukit.core.dataset import Dataset, load_dataset
from plateaukit.installer import (
    install_dataset,
    list_installed_datasets,
    uninstall_dataset,
)

__all__ = [
    "Dataset",
    "load_dataset",
    "install_dataset",
    "uninstall_dataset",
    "list_installed_datasets",
]



from lxml import etree

from plateaukit.formats.citygml.constants import nsmap


class CodelistParser:
    """A parser for codelists."""

    def __init__(self):
        pass

    def parse(self, infile):
        tree = etree.parse(infile)
        root = tree.getroot()

        result = dict()

        for el in root.iterfind(".//gml:dictionaryEntry/gml:Definition", nsmap):
            # print(el)

            el_key = el.find("./gml:name", nsmap)
            key = el_key.text if el_key is not None else None

            el_value = el.find("./gml:description", nsmap)
            value = el_value.text if el_value is not None else None

            if key is not None:
                result[key] = value

        return result



from io import IOBase
import re
from typing import BinaryIO

import pyproj
from lxml import etree

from plateaukit.formats.citygml import CityGML
from plateaukit.formats.citygml.constants import nsmap
from plateaukit.formats.citygml.parsers.city_object_parser import (
    PLATEAUCityObjectParser,
)
from plateaukit.formats.citygml.parsers.codelist_parser import CodelistParser


class CityGMLParser:
    pass


class PLATEAUCityGMLParser(CityGMLParser):
    """A parser for PLATEAU CityGML.

    Attributes:
        target_epsg: Target EPSG code.
        codelist_file_map: A map from codelist path to file object.
    """

    target_epsg: int
    codelist_file_map: dict[str, BinaryIO] | None = None

    def __init__(self, target_epsg: int = 4326, codelist_file_map=None):
        self.target_epsg = target_epsg
        self.codelist_file_map = codelist_file_map

    def _get_epsg_code(self, infile: IOBase) -> str | None:
        """Extract EPSG code from CityGML tree."""
        tag = f"{{{nsmap['gml']}}}boundedBy"
        path = "gml:Envelope"

        srs_name = None
        for _ev, el in etree.iterparse(infile, events=("end",), tag=tag):
            result = el.find(path, nsmap)
            found = result is not None
            if result is not None:
                srs_name = result.get("srsName")
                break
            el.clear()

        if srs_name is None:
            raise Exception("EPSG code not found")

        m = re.match(r"http://www.opengis.net/def/crs/EPSG/0/(\d+)", srs_name)

        if not m:
            raise Exception("Failed to parse EPSG code")

        infile.seek(0)

        return m.group(1)

    def iterparse(self, infile, selection: list[str] | None = None):
        src_epsg = self._get_epsg_code(infile)  # 6697

        # TODO: Accept options like always_xy
        transformer = pyproj.Transformer.from_crs(src_epsg, self.target_epsg)

        # Parse codelists
        codelist_map = None

        if self.codelist_file_map:
            codelist_map = dict()

            parser = CodelistParser()

            for path, file_obj in self.codelist_file_map.items():
                codelist = parser.parse(file_obj)
                codelist_map[path] = codelist

        co_parser = PLATEAUCityObjectParser(
            transformer=transformer, codelist_map=codelist_map
        )

        objects = []

        tag = f"{{{nsmap['core']}}}cityObjectMember"

        itertree = etree.iterparse(infile, events=("end",), tag=tag)
        _, root = next(itertree)

        for _ev, el in itertree:
            it = el.iterchildren()
            co_element = next(it)
            obj = co_parser.parse(co_element)

            # TODO: Improve performance
            building_id = (
                obj.attributes.get("building_id", None) if obj.attributes else None
            )
            if selection and (
                obj.id not in selection and (building_id not in selection)
            ):
                continue

            yield obj

            el.clear()
            root.clear()

        infile.seek(0)

    def parse(self, infile, selection: list[str] | None = None):
        src_epsg = self._get_epsg_code(infile)  # 6697

        # TODO: Accept options like always_xy
        transformer = pyproj.Transformer.from_crs(src_epsg, self.target_epsg)

        # Parse codelists
        codelist_map = None

        if self.codelist_file_map:
            codelist_map = dict()

            parser = CodelistParser()

            for path, file_obj in self.codelist_file_map.items():
                codelist = parser.parse(file_obj)
                codelist_map[path] = codelist

        co_parser = PLATEAUCityObjectParser(
            transformer=transformer, codelist_map=codelist_map
        )

        objects = []

        tag = f"{{{nsmap['core']}}}cityObjectMember"

        itertree = etree.iterparse(infile, events=("end",), tag=tag)
        _, root = next(itertree)

        for _ev, el in itertree:
            it = el.iterchildren()
            co_element = next(it)
            obj = co_parser.parse(co_element)

            # TODO: Improve performance
            building_id = (
                obj.attributes.get("building_id", None) if obj.attributes else None
            )
            if selection and (
                obj.id not in selection and (building_id not in selection)
            ):
                continue

            # TODO: Memory optimization
            objects.append(obj)

            el.clear()
            root.clear()

        infile.seek(0)

        return CityGML(city_objects=objects)



from plateaukit.constants import nsmap

# NOTE: https://www.mlit.go.jp/plateau/file/libraries/doc/plateau_doc_0001_ver02.pdf#page=336
data_type_display_names = {
    "bldg": {
        "ja": "建築物",
    },
    # NOTE: brid is not defined in the specification
    "brid": {
        "ja": "橋梁",
    },
    "tran": {
        "ja": "道路",
    },
    "urf": {
        "ja": "都市計画決定情報",
    },
    "luse": {
        "ja": "土地利用",
    },
    "fld": {
        "ja": "洪水浸水想定区域",
    },
    "tnm": {
        "ja": "津波浸水想定",
    },
    "lsld": {
        "ja": "土砂災害警戒区域",
    },
    "htd": {
        "ja": "高潮浸水想定区域",
    },
    "ifld": {
        "ja": "内水浸水想定区域",
    },
    "frn": {
        "ja": "都市設備",
    },
    "veg": {
        "ja": "植生",
    },
    "dem": {
        "ja": "起伏",
    },
}

tag_display_names = {
    "gml:name": {
        "ja": "名称",
    },
    "gml:description": {
        "ja": "説明",
    },
    "gml:boundedBy": {
        "ja": "範囲",
    },
    "core:creationDate": {
        "ja": "作成日",
    },
    "bldg:class": {
        "ja": "分類",
    },
    "bldg:usage": {
        "ja": "用途",
    },
    "bldg:measuredHeight": {
        "ja": "計測高さ",
    },
    "bldg:storeysAboveGround": {
        "ja": "地上階数",
    },
    "bldg:lod0RoofEdge": {
        "ja": "lod0屋根面",
    },
    "bldg:lod0FootPrint": {
        "ja": "lod0接地面",
    },
    "bldg:lod1Solid": {
        "ja": "lod1立体",
    },
    "bldg:lod2Solid": {
        "ja": "lod2立体",
    },
    "bldg:lod3Solid": {
        "ja": "lod3立体",
    },
    "bldg:address": {
        "ja": "住所",
    },
    "bldg:boundedBy": {
        "ja": "境界面",
    },
    "bldg:outerBuildingInstallation": {
        "ja": "建物付属物",
    },
    "bldg:consistsOfBuildingPart": {
        "ja": "建物部品",
    },
    "tran:function": {
        "ja": "機能",
    },
    "dem:reliefComponent": {
        "ja": "地形構成要素",
    },
    "uro:buildingDataQualityAttribute": {
        "ja": "データ品質",
    },
    "uro:buildingDetailAttribute": {
        "ja": "建物利用現況",
    },
    "uro:buildingIDAttribute": {
        "ja": "建物識別情報",
    },
    "uro:buildingDisasterRiskAttribute": {
        "ja": "災害リスク",
    },
}



import concurrent.futures
import math
from multiprocessing import Manager
from pathlib import Path

from rich.progress import Progress

from plateaukit import parallel, utils
from plateaukit.logger import logger

from ._serial import cityson_from_gml_serial_with_quit


def cityjson_from_citygml(
    infiles,
    outfile,
    *,
    split: int = 1,
    zipfile=None,
    precision=16,
    object_types=None,
    lod: list[int] = [1, 2],
    ground: bool = False,
    seq=False,
    codelist_infiles=None,
    selection: list[str] | None = None,
    target_epsg: int | None = None,
    progress={},
):
    logger.debug("[*] cityjson_from_citygml")
    # vertices_map = VerticesMap()
    # city_objects = {}

    group_size = math.ceil(len(infiles) / split)
    logger.debug(f"GMLs per GeoJSON: {group_size}")
    infile_groups = utils.chunker(infiles, group_size)

    with Progress() as rprogress:
        overall_progress_description = progress.get("description", "Processing...")
        overall_task_id = rprogress.add_task(overall_progress_description)

        with Manager() as manager:
            quit = manager.Event()
            _progress = manager.dict()

            with concurrent.futures.ProcessPoolExecutor(max_workers=None) as pool:
                futures = []
                futures_status = dict()
                for i, infile_group in enumerate(infile_groups):
                    stem = Path(outfile).stem

                    if split > 1:
                        group_outfile = Path(outfile).with_stem(f"{stem}.{i + 1}")
                    else:
                        group_outfile = outfile

                    task_id = rprogress.add_task(
                        f"[cyan]Progress #{i + 1}", total=len(infile_group)
                    )
                    # task_id = None

                    logger.debug(f"group_outfile: {group_outfile}")

                    future = pool.submit(
                        cityson_from_gml_serial_with_quit,
                        infile_group,
                        group_outfile,
                        object_types=object_types,
                        lod=lod,
                        ground=ground,
                        codelist_infiles=codelist_infiles,
                        seq=seq,
                        zipfile=zipfile,
                        selection=selection,
                        target_epsg=target_epsg,
                        task_id=task_id,
                        _progress=_progress,
                        quit=quit,
                        # **kwargs,
                    )

                    futures.append(future)
                    futures_status[future] = {
                        "task_id": task_id,
                        "counter": i + 1,
                        "failed": False,
                    }

                parallel.wait_futures(
                    futures,
                    pool,
                    quit,
                    futures_status,
                    overall_progress={
                        "task_id": overall_task_id,
                        "description": overall_progress_description,
                    },
                    rich_progress=rprogress,
                    shared_progress_status=_progress,
                )



from dataclasses import dataclass
from typing import Any


@dataclass
class CityObject:
    """A CityGML CityObject."""

    type: str
    id: str | None = None
    attributes: dict[str, Any] | None = None
    geometry: list[Any] | None = None
    # children: list[Self] | None = None
    # parent: Self | None = None


# TODO: boundedBy



import os.path
import shutil
from pathlib import Path
from urllib.parse import urlparse

import requests
from rich.progress import wrap_file

API_URL = "https://www.geospatial.jp/ckan/api/3"


def download_resource(resource_id, dest="/tmp"):
    resp = requests.get(f"{API_URL}/action/resource_show", params={"id": resource_id})
    data = resp.json()

    if not data["success"]:
        raise Exception("Failed to download")

    file_url = data["result"]["url"]
    file_name = os.path.basename(urlparse(file_url).path)

    destfile_path = str(Path(dest, file_name))

    # NOTE: If the destfile_path is in the home directory, replace it with ~ for better privacy
    try:
        _destfile_path = Path(destfile_path).relative_to(Path.home())
        _destfile_path = str(Path("~", _destfile_path))
    except:
        _destfile_path = destfile_path

    print(f"Download file as: {_destfile_path}")

    with requests.get(file_url, stream=True) as r:
        value = r.headers.get("Content-Length")
        total_length = int(value) if value is not None else -1
        with wrap_file(r.raw, total=total_length, description="Downloading...") as raw:
            with open(destfile_path, "wb") as output:
                shutil.copyfileobj(raw, output)

    return destfile_path


# download_resource("0bab2b7f-6962-41c8-872f-66ad9b40dcb1")

# NOTE: format
# {
#     "help": "https://www.geospatial.jp/ckan/api/3/action/help_show?name=resource_show",
#     "success": True,
#     "result": {
#         "cache_last_updated": None,
#         "standard_price": "",
#         "package_id": "b7a9b937-2e68-4f33-9c4e-16b5cc7392a8",
#         "datastore_active": False,
#         "id": "0bab2b7f-6962-41c8-872f-66ad9b40dcb1",
#         "size": None,
#         "tos": "",
#         "selection_type": "",
#         "state": "active",
#         "hash": "",
#         "description": "CityGML形式のデータで、次のデータが格納されています。\r\n\r\n* 建築物\r\n* 橋梁\r\n* 道路\r\n* 土地利用・公園\r\n* 地形\r\n* 都市設備\r\n* 洪水浸水想定区域（国管理、都管理）\r\n* 土砂災害警戒区域\r\n* 製品仕様書（PDF形式）\r\n* コードインデックス（XML形式）\r\n* 構築範囲図（PDF形式）\r\n* メタデータ（XML形式）\r\n\r\nファイルサイズは約4.2GBです。\r\n\r\n※2021年9月18日\u3000i-UR1.4は、名前空間及びXMLSchemaファイルの所在が変更されたことに伴い、i-UR1.5に改定されています。\r\n[(https://www.chisou.go.jp/tiiki/toshisaisei/itoshisaisei/iur)](https://www.chisou.go.jp/tiiki/toshisaisei/itoshisaisei/iur)\r\n\r\n各ファイルにおけるこれらの記述は、改定前の記述となっていますのでご注意ください。\r\n\r\n** 更新情報 **\r\n\r\n2022年4月1日\u3000CityGMLのディレクトリを修正しました。\r\n",
#         "format": "ZIP",
#         "last_modified": None,
#         "url_type": None,
#         "metadata_type": "",
#         "mimetype": None,
#         "cache_url": None,
#         "name": "CityGML",
#         "acknowledgement": "",
#         "url": "https://gic-plateau.s3.ap-northeast-1.amazonaws.com/2020/13100_tokyo23-ku_2020_citygml_3_2_op.zip",
#         "created": "2022-04-01T14:31:28.753939",
#         "mimetype_inner": None,
#         "position": 2,
#         "revision_id": "df6da242-128e-42a0-93e4-7b08c607de5d",
#         "data_crs": "",
#         "resource_type": None,
#     },
# }



from __future__ import annotations

import click

from plateaukit import generators
from plateaukit.core.dataset import load_dataset


@click.command("generate-cityjson")
@click.argument("infiles", nargs=-1)
@click.argument("outfile", nargs=1, required=True)
@click.option("--dataset", "dataset_id", help='Dataset ID (e.g. "plateau-tokyo23ku")')
@click.option(
    "--type",
    "-t",
    "types",
    type=click.Choice(
        ["bldg", "brid", "dem", "fld", "frn", "lsld", "luse", "tran", "urf"],
        case_sensitive=True,
    ),
    default=["bldg"],
    multiple=True,
)
@click.option(
    "--split", default=1, help="Split the output into specified number of files"
)
@click.option(
    "--ground", is_flag=True, default=False, help="Shift objects to the ground level"
)
@click.option("--seq", is_flag=True, default=False, help="Generate CityJSONSeq")
@click.option(
    "--target-epsg", default=None, help="EPSG code for the output CityJSON file"
)
# # @click.option(
# #     "--precision",
# #     help="Number of decimal places to keep for geometry vertices (default: 16).",
# )
def generate_cityjson_cmd(
    infiles, outfile, dataset_id, types, split, ground, seq, target_epsg
):
    """Generate CityJSON from PLATEAU datasets.

    PLATEAU データセットから CityJSON を生成します。
    """

    if not infiles and not dataset_id:
        raise click.UsageError("Missing argument/option: infiles or --dataset")

    if infiles and dataset_id:
        raise click.UsageError("Too many argument")

    params = {}

    # if precision:
    #     params["precision"] = precision

    # obj_types_map = {
    #     "bldg": ["Building"],
    #     "brid": ["Bridge"],
    #     "tran": ["Road"],
    #     # "dem": ["ReliefFeature"],
    #     # "fld": ["LandUse"],
    #     # "frn": ["WaterBody"],
    #     # "lsld": ["TransportationComplex"],
    #     # "luse": ["LandUse"],
    #     # "urf": ["CityFurniture"],
    # }

    # obj_types = sum([obj_types_map[type] for type in types], [])
    # logger.debug(obj_types)

    if dataset_id:
        dataset = load_dataset(dataset_id)
        dataset.to_cityjson(
            outfile,
            types=types,
            split=split,
            seq=seq,
            ground=ground,
            target_epsg=target_epsg,
            **params,
        )

    else:
        generators.cityjson.cityjson_from_citygml(infiles, outfile, **params)


def _generate_geojson(
    infiles, outfile, dataset_id: str, types: list[str], seq: bool, split: int, **kwargs
):
    """Generate GeoJSON from PLATEAU datasets."""

    if not infiles and not dataset_id:
        raise click.UsageError("Missing argument: infiles or dataset")

    if infiles and dataset_id:
        raise click.UsageError("Too many arguments")

    if dataset_id:
        dataset = load_dataset(dataset_id)
        dataset.to_geojson(outfile, types=types, seq=seq, split=split, **kwargs)
    else:
        generators.geojson.geojson_from_citygml(
            infiles, outfile, types=types, seq=seq, split=split, **kwargs
        )


@click.command("generate-geojson")
@click.argument("infiles", nargs=-1)
@click.argument("outfile", nargs=1, required=True)
@click.option("--dataset", "dataset_id", help='Dataset ID (e.g. "plateau-tokyo23ku")')
@click.option(
    "--type",
    "-t",
    "types",
    type=click.Choice(
        ["bldg", "brid", "dem", "fld", "frn", "lsld", "luse", "tran", "urf"],
        case_sensitive=True,
    ),
    default=["bldg"],
    multiple=True,
)
@click.option("--seq", is_flag=True, default=False, help="Generate GeoJSONSeq")
@click.option("--split", default=1)
def generate_geojson_cmd(
    infiles, outfile, dataset_id: str, types: list[str], seq: bool, split: int
):
    """Generate GeoJSON from PLATEAU datasets.

    PLATEAU データセットから GeoJSON を生成します。
    """

    _generate_geojson(infiles, outfile, dataset_id, types=types, seq=seq, split=split)


@click.command("generate-qmesh")
@click.argument("infiles", nargs=-1)
@click.argument("outfile", nargs=1, required=True)
def generate_qmesh_cmd(infiles, outfile):
    """Generate Quantized Mesh from PLATEAU datasets. (Requires `pip install 'plateaukit[quantized_mesh]'`)

    PLATEAU データセットから Quantized Mesh を生成します。(要 `pip install 'plateaukit[quantized_mesh]'`)
    """

    generators.triangles_from_gml(infiles)


# @cli.command("generate-gpkg")
# @click.argument("infiles", nargs=-1, required=True)
# @click.argument("outfile", nargs=1, required=True)
# def generate_gpkg(infiles, outfile):
#     """Generate GeoPackage from PLATEAU GeoJSON."""
#     expanded_infiles = []
#     for infile in infiles:
#         expanded_infiles.extend(glob.glob(infile))
#     generators.utils.geojson_to_gpkg(expanded_infiles, outfile)


# @cli.command("generate-heightmap")
# @click.argument("infiles", nargs=-1)
# @click.argument("outfile", nargs=1, required=True)
# def generate_heightmap(infiles, outfile):
#     """Generate GeoTIFF heightmap from PLATEAU CityGML."""
#     expanded_infiles = []
#     for infile in infiles:
#         expanded_infiles.extend(glob.glob(infile))
#     generators.triangles_from_gml(expanded_infiles)


# @cli.command("extract-properties")
# @click.argument("infiles", nargs=-1, required=True)
# @click.argument("outfile", nargs=1, required=True)
# def extract_properties(infiles, outfile):
#     """Extract properties from PLATEAU CityGML."""
#     expanded_infiles = []
#     for infile in infiles:
#         expanded_infiles.extend(glob.glob(infile))
#     run_async(extractors.commands.extract_properties(expanded_infiles, outfile))



from __future__ import annotations

import click

from plateaukit.config import Config


@click.command("config")
def config_cmd():
    """Show PlateauKit configuration information.

    PlateauKit の設定情報を表示します。
    """
    import importlib.metadata
    import json

    try:
        __version__ = importlib.metadata.version("plateaukit")
    except importlib.metadata.PackageNotFoundError:
        __version__ = "unknown"

    config = Config()
    click.echo(f"Version: {__version__}")
    click.echo(f"Config path: {config.path}")
    click.echo(f"Data directory: {config.data_dir}")
    click.echo(f"{json.dumps(config.datasets, indent=2, ensure_ascii=False)}")



import warnings

from lxml import etree

from plateaukit.formats.citygml.constants import nsmap


# TODO: Inherit etree.Element?
# NOTE: This class MUST be lightweight
class CityObjectXML:
    """A CityGML object XML with utility methods.

    Attributes:
        codelist_map: A map from codelist path to corresponding dict.
    """

    codelist_map: dict[str, dict[str, str]]  # TODO: Class or instance variable?

    def __init__(
        self, el: etree._Element, codelist_map: dict[str, dict[str, str]] = {}
    ):
        self.tree = el
        self.codelist_map = codelist_map

    def _get_codespace_attribute(self, xpath) -> str | None:
        el = self.tree.find(xpath, nsmap)

        if el is None:
            return None

        # Check if el has codeSpace attribute
        if (code_space_path := el.get("codeSpace")) is not None:
            code_dict = self.codelist_map.get(code_space_path, None)
            key = el.text

            if code_dict is None or key is None:
                return None

            value = code_dict.get(key, None)
            return value

        value = el.text if el is not None else None
        return value

    def get_gml_id(self) -> str | None:
        """Get gml:id of a CityGML object."""

        path = "./[@gml:id]"
        result = self.tree.find(path, nsmap)

        if result is None:
            warnings.warn(
                "gml:id not found"
                # f"gml:id not found\n{etree.tostring(tree, pretty_print=True).decode()}"
            )
            return None

        id = result.get(f"{{{nsmap['gml']}}}id")

        return id if id is not None else None

    def find(self, *args, **kwargs):
        return self.tree.find(*args, **kwargs)
